@@ -2,7 +2,7 @@
  * colorize - Read text from standard input stream or file and print
  *            it colorized through use of ANSI escape sequences
  *
- * Copyright (c) 2011-2017 Steven Schubiger
+ * Copyright (c) 2011-2018 Steven Schubiger
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -73,7 +73,7 @@
 #define LF 0x01
 #define CR 0x02
 
-#define SKIP_LINE_ENDINGS(flags) (((flags) & CR) && ((flags) & LF) ? 2 : 1)
+#define SKIP_LINE_ENDINGS(flags) ((flags) == (CR|LF) ? 2 : 1)
 
 #define VALID_FILE_TYPE(mode) (S_ISREG (mode) || S_ISLNK (mode) || S_ISFIFO (mode))
 
@@ -99,12 +99,12 @@
 
 #define ABORT_TRACE()                                                              \
     fprintf (stderr, "Aborting in source file %s, line %u\n", __FILE__, __LINE__); \
-    abort ();                                                                      \
+    abort ();
 
 #define CHECK_COLORS_RANDOM(color1, color2)        \
      streq (color_names[color1]->name, "random")   \
  && (streq (color_names[color2]->name, "none")     \
-  || streq (color_names[color2]->name, "default")) \
+  || streq (color_names[color2]->name, "default"))
 
 #define ALLOC_COMPLETE_PART_LINE 8
 
@@ -116,11 +116,15 @@
 # define COLOR_SEP_CHAR '/'
 #endif
 
-#define DEBUG_FILE "debug.txt"
+#if DEBUG
+# define DEBUG_FILE "debug.txt"
+#endif
 
-#define MAX_ATTRIBUTE_CHARS (5 * 2)
+#define MAX_ATTRIBUTE_CHARS (6 * 2)
 
-#define VERSION "0.61"
+#define PROGRAM_NAME "colorize"
+
+#define VERSION "0.64"
 
 typedef enum { false, true } bool;
 
@@ -164,7 +168,7 @@ struct bytes_size {
     char unit;
 };
 
-enum fmts {
+enum {
     FMT_GENERIC,
     FMT_STRING,
     FMT_QUOTE,
@@ -188,7 +192,7 @@ static const char *formats[] = {
 enum { GENERIC, FOREGROUND = 0, BACKGROUND };
 
 static const struct {
-    struct color const *entries;
+    const struct color *entries;
     unsigned int count;
     const char *desc;
 } tables[] = {
@@ -201,18 +205,33 @@ enum {
     OPT_CLEAN,
     OPT_CLEAN_ALL,
     OPT_EXCLUDE_RANDOM,
+    OPT_OMIT_COLOR_EMPTY,
     OPT_HELP,
     OPT_VERSION
 };
 static int opt_type;
 static const struct option long_opts[] = {
-    { "attr",           required_argument, &opt_type, OPT_ATTR           },
-    { "clean",          no_argument,       &opt_type, OPT_CLEAN          },
-    { "clean-all",      no_argument,       &opt_type, OPT_CLEAN_ALL      },
-    { "exclude-random", required_argument, &opt_type, OPT_EXCLUDE_RANDOM },
-    { "help",           no_argument,       &opt_type, OPT_HELP           },
-    { "version",        no_argument,       &opt_type, OPT_VERSION        },
-    {  NULL,            0,                 NULL,      0                  },
+    { "attr",             required_argument, &opt_type, OPT_ATTR             },
+    { "clean",            no_argument,       &opt_type, OPT_CLEAN            },
+    { "clean-all",        no_argument,       &opt_type, OPT_CLEAN_ALL        },
+    { "exclude-random",   required_argument, &opt_type, OPT_EXCLUDE_RANDOM   },
+    { "omit-color-empty", no_argument,       &opt_type, OPT_OMIT_COLOR_EMPTY },
+    { "help",             no_argument,       &opt_type, OPT_HELP             },
+    { "version",          no_argument,       &opt_type, OPT_VERSION          },
+    {  NULL,              0,                 NULL,      0                    },
+};
+
+enum attr_type {
+    ATTR_BOLD = 0x01,
+    ATTR_UNDERSCORE = 0x02,
+    ATTR_BLINK = 0x04,
+    ATTR_REVERSE = 0x08,
+    ATTR_CONCEALED = 0x10
+};
+struct attr {
+    const char *name;
+    unsigned int val;
+    enum attr_type type;
 };
 
 static FILE *stream;
@@ -225,6 +244,7 @@ static void **vars_list;
 
 static bool clean;
 static bool clean_all;
+static bool omit_color_empty;
 
 static char attr[MAX_ATTRIBUTE_CHARS + 1];
 static char *exclude;
@@ -233,7 +253,7 @@ static const char *program_name;
 
 static void process_opts (int, char **);
 static void process_opt_attr (const char *);
-static void write_attr (unsigned int);
+static void write_attr (const struct attr *, unsigned int *);
 static void print_hint (void);
 static void print_help (void);
 static void print_version (void);
@@ -250,7 +270,7 @@ static bool get_next_char (char *, const char **, FILE *, bool *);
 static void save_char (char, char **, size_t *, size_t *);
 static void find_color_entries (struct color_name **, const struct color **);
 static void find_color_entry (const struct color_name *, unsigned int, const struct color **);
-static void print_line (const char *, const struct color **, const char * const, unsigned int);
+static void print_line (const char *, const struct color **, const char * const, unsigned int, bool);
 static void print_clean (const char *);
 static bool is_esc (const char *);
 static const char *get_end_of_esc (const char *);
@@ -319,8 +339,8 @@ main (int argc, char **argv)
           vfprintf_fail (formats[FMT_GENERIC], "--clean and --clean-all switch are mutually exclusive");
         if (arg_cnt > 1)
           {
-            const char *format = "%s %s";
-            const char *message = "switch cannot be used with more than one file";
+            const char *const format = "%s %s";
+            const char *const message = "switch cannot be used with more than one file";
             if (clean)
               vfprintf_fail (format, "--clean", message);
             else if (clean_all)
@@ -350,11 +370,11 @@ main (int argc, char **argv)
 
 #define PRINT_HELP_EXIT() \
     print_help ();        \
-    exit (EXIT_SUCCESS);  \
+    exit (EXIT_SUCCESS);
 
 #define PRINT_VERSION_EXIT() \
     print_version ();        \
-    exit (EXIT_SUCCESS);     \
+    exit (EXIT_SUCCESS);
 
 extern char *optarg;
 
@@ -396,6 +416,9 @@ process_opts (int argc, char **argv)
                       vfprintf_fail (formats[FMT_GENERIC], "--exclude-random switch must be provided a plain color");
                     break;
                   }
+                  case OPT_OMIT_COLOR_EMPTY:
+                    omit_color_empty = true;
+                    break;
                   case OPT_HELP:
                     PRINT_HELP_EXIT ();
                   case OPT_VERSION:
@@ -420,6 +443,16 @@ process_opts (int argc, char **argv)
 static void
 process_opt_attr (const char *p)
 {
+    /* If attributes are added to this "list", also increase MAX_ATTRIBUTE_CHARS!  */
+    const struct attr attrs[] = {
+        { "bold",       1, ATTR_BOLD       },
+        { "underscore", 4, ATTR_UNDERSCORE },
+        { "blink",      5, ATTR_BLINK      },
+        { "reverse",    7, ATTR_REVERSE    },
+        { "concealed",  8, ATTR_CONCEALED  },
+    };
+    unsigned int attr_types = 0;
+
     while (*p)
       {
         const char *s;
@@ -432,19 +465,27 @@ process_opt_attr (const char *p)
           vfprintf_fail (formats[FMT_GENERIC], "--attr switch must have strings separated by ,");
         else
           {
-            /* If attributes are added to this "list", also increase MAX_ATTRIBUTE_CHARS!  */
-            if (p - s == 4 && strneq (s, "bold", 4))
-              write_attr (1);
-            else if (p - s == 10 && strneq (s, "underscore", 10))
-              write_attr (4);
-            else if (p - s == 5 && strneq (s, "blink", 5))
-              write_attr (5);
-            else if (p - s == 7 && strneq (s, "reverse", 7))
-              write_attr (7);
-            else if (p - s == 9 && strneq (s, "concealed", 9))
-              write_attr (8);
-            else
-              vfprintf_fail (formats[FMT_GENERIC], "--attr switch must be provided valid attribute names");
+            bool valid_attr = false;
+            unsigned int i;
+            for (i = 0; i < sizeof (attrs) / sizeof (struct attr); i++)
+              {
+                const size_t name_len = strlen (attrs[i].name);
+                if ((size_t)(p - s) == name_len && strneq (s, attrs[i].name, name_len))
+                  {
+                    write_attr (&attrs[i], &attr_types);
+                    valid_attr = true;
+                    break;
+                  }
+              }
+            if (!valid_attr)
+              {
+                char *attr_invalid = xmalloc ((p - s) + 1);
+                STACK_VAR (attr_invalid);
+                strncpy (attr_invalid, s, p - s);
+                attr_invalid[p - s] = '\0';
+                vfprintf_fail ("--attr switch attribute '%s' is not valid", attr_invalid);
+                RELEASE_VAR (attr_invalid); /* never reached */
+              }
           }
         if (*p)
           p++;
@@ -452,9 +493,16 @@ process_opt_attr (const char *p)
 }
 
 static void
-write_attr (unsigned int val)
+write_attr (const struct attr *attr_i, unsigned int *attr_types)
 {
+    const unsigned int val = attr_i->val;
+    const enum attr_type attr_type = attr_i->type;
+    const char *attr_name = attr_i->name;
+
+    if (*attr_types & attr_type)
+      vfprintf_fail ("--attr switch has attribute '%s' twice or more", attr_name);
     snprintf (attr + strlen (attr), 3, "%u;", val);
+    *attr_types |= attr_type;
 }
 
 static void
@@ -466,13 +514,16 @@ print_hint (void)
 static void
 print_help (void)
 {
-    struct short_opt {
+    struct opt_data {
         const char *name;
         const char *short_opt;
+        const char *arg;
     };
-    const struct short_opt short_opts[] = {
-        { "help",    "h" },
-        { "version", "V" },
+    const struct opt_data opts_data[] = {
+        { "attr",           NULL, "=ATTR1,ATTR2,..." },
+        { "exclude-random", NULL, "=COLOR"           },
+        { "help",           "h",  NULL               },
+        { "version",        "V",  NULL               },
     };
     const struct option *opt = long_opts;
     unsigned int i;
@@ -498,18 +549,21 @@ print_help (void)
     printf ("\n\tOptions\n");
     for (; opt->name; opt++)
       {
-        const char *short_opt = NULL;
+        const struct opt_data *opt_data = NULL;
         unsigned int i;
-        for (i = 0; i < sizeof (short_opts) / sizeof (struct short_opt); i++)
+        for (i = 0; i < sizeof (opts_data) / sizeof (struct opt_data); i++)
+          if (streq (opt->name, opts_data[i].name))
+            {
+              opt_data = &opts_data[i];
+              break;
+            }
+        if (opt_data)
           {
-            if (streq (opt->name, short_opts[i].name))
-              {
-                short_opt = short_opts[i].short_opt;
-                break;
-              }
+            if (opt_data->short_opt)
+              printf ("\t\t-%s, --%s\n", opt_data->short_opt, opt->name);
+            else
+              printf ("\t\t    --%s%s\n", opt->name, opt_data->arg);
           }
-        if (short_opt)
-          printf ("\t\t-%s, --%s\n", short_opt, opt->name);
         else
           printf ("\t\t    --%s\n", opt->name);
       }
@@ -522,26 +576,27 @@ print_version (void)
 #ifdef HAVE_VERSION
 # include "version.h"
 #else
-    const char *version = NULL;
+    const char *const version = NULL;
 #endif
     const char *version_prefix, *version_string;
     const char *c_flags, *ld_flags, *cpp_flags;
+    const char *const desc_flags_unknown = "unknown";
     struct bytes_size bytes_size;
     bool debug;
 #ifdef CFLAGS
     c_flags = to_str (CFLAGS);
 #else
-    c_flags = "unknown";
+    c_flags = desc_flags_unknown;
 #endif
 #ifdef LDFLAGS
     ld_flags = to_str (LDFLAGS);
 #else
-    ld_flags = "unknown";
+    ld_flags = desc_flags_unknown;
 #endif
 #ifdef CPPFLAGS
     cpp_flags = to_str (CPPFLAGS);
 #else
-    cpp_flags = "unknown";
+    cpp_flags = desc_flags_unknown;
 #endif
 #if DEBUG
     debug = true;
@@ -550,7 +605,7 @@ print_version (void)
 #endif
     version_prefix = version ? "" : "v";
     version_string = version ? version : VERSION;
-    printf ("colorize %s%s (compiled at %s, %s)\n", version_prefix, version_string, __DATE__, __TIME__);
+    printf ("%s %s%s (compiled at %s, %s)\n", PROGRAM_NAME, version_prefix, version_string, __DATE__, __TIME__);
 
     printf ("Compiler flags: %s\n", c_flags);
     printf ("Linker flags: %s\n", ld_flags);
@@ -606,12 +661,16 @@ process_args (unsigned int arg_cnt, char **arg_strings, char *attr, const struct
     int ret;
     char *p;
     struct stat sb;
-    struct color_name *color_names[3] = { NULL, NULL, NULL };
+    struct color_name *color_names[3] = {
+        NULL, /* foreground */
+        NULL, /* background */
+        NULL, /* sentinel value */
+    };
 
     const char *color_string = arg_cnt >= 1 ? arg_strings[0] : NULL;
     const char *file_string  = arg_cnt == 2 ? arg_strings[1] : NULL;
 
-    assert (color_string);
+    assert (color_string != NULL);
 
     if (streq (color_string, "-"))
       {
@@ -640,7 +699,7 @@ process_args (unsigned int arg_cnt, char **arg_strings, char *attr, const struct
 
     gather_color_names (color_string, attr, color_names);
 
-    assert (color_names[FOREGROUND]);
+    assert (color_names[FOREGROUND] != NULL);
 
     if (color_names[BACKGROUND])
       {
@@ -656,6 +715,7 @@ process_args (unsigned int arg_cnt, char **arg_strings, char *attr, const struct
       }
 
     find_color_entries (color_names, colors);
+    assert (colors[FOREGROUND] != NULL);
     free_color_names (color_names);
 
     if (!colors[FOREGROUND]->code && colors[BACKGROUND] && colors[BACKGROUND]->code)
@@ -664,6 +724,7 @@ process_args (unsigned int arg_cnt, char **arg_strings, char *attr, const struct
         color_name.name = color_name.orig = "default";
 
         find_color_entry (&color_name, FOREGROUND, colors);
+        assert (colors[FOREGROUND]->code != NULL);
       }
 
     process_file_arg (file_string, file, stream);
@@ -701,8 +762,8 @@ process_file_arg (const char *file_string, const char **file, FILE **stream)
         *file = "stdin";
       }
 
-    assert (*stream);
-    assert (*file);
+    assert (*stream != NULL);
+    assert (*file != NULL);
 }
 
 static void
@@ -742,15 +803,15 @@ skip_path_colors (const char *color_string, const char *file_string, const struc
 
     if (have_file)
       {
-        const char *file_exists = color_string;
+        const char *file_existing = color_string;
         if (file_string)
-          vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_exists, "cannot be used as color string");
+          vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_existing, "cannot be used as color string");
         else
           {
             if (VALID_FILE_TYPE (mode))
-              vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_exists, "must be preceded by color string");
+              vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_existing, "must be preceded by color string");
             else
-              vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_exists, "is not a valid file type");
+              vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_existing, "is not a valid file type");
           }
       }
 }
@@ -776,7 +837,7 @@ gather_color_names (const char *color_string, char *attr, struct color_name **co
           }
         else
           p = color + strlen (color);
-        assert (p);
+        assert (p != NULL);
 
         for (ch = color; *ch; ch++)
           if (!isalpha (*ch))
@@ -837,6 +898,7 @@ read_print_stream (const char *attr, const struct color **colors, const char *fi
         line = buf;
         while ((eol = strpbrk (line, "\n\r")))
           {
+            const bool has_text = (eol > line);
             const char *p;
             flags &= ~(CR|LF);
             if (*eol == '\r')
@@ -847,17 +909,18 @@ read_print_stream (const char *attr, const struct color **colors, const char *fi
               }
             else if (*eol == '\n')
               flags |= LF;
-            else
+            else /* never reached */
               vfprintf_fail (formats[FMT_FILE], file, "unrecognized line ending");
             p = eol + SKIP_LINE_ENDINGS (flags);
             *eol = '\0';
-            print_line (attr, colors, line, flags);
+            print_line (attr, colors, line, flags,
+                        omit_color_empty ? has_text : true);
             line = p;
           }
         if (feof (stream))
           {
             if (*line != '\0')
-              print_line (attr, colors, line, 0);
+              print_line (attr, colors, line, 0, true);
           }
         else if (*line != '\0')
           {
@@ -865,7 +928,7 @@ read_print_stream (const char *attr, const struct color **colors, const char *fi
             if ((clean || clean_all) && (p = strrchr (line, '\033')))
               merge_print_line (line, p, stream);
             else
-              print_line (attr, colors, line, 0);
+              print_line (attr, colors, line, 0, true);
           }
       }
 }
@@ -932,7 +995,7 @@ complete_part_line (const char *p, char **buf, FILE *stream)
             if (read_from_stream)
               save_char (ch, buf, &i, &size);
           }
-        else /* read next character */
+        else /* got next character */
           {
             got_next_char = true;
             break;
@@ -1072,12 +1135,13 @@ find_color_entry (const struct color_name *color_name, unsigned int index, const
 }
 
 static void
-print_line (const char *attr, const struct color **colors, const char *const line, unsigned int flags)
+print_line (const char *attr, const struct color **colors, const char *const line, unsigned int flags, bool emit_colors)
 {
     /* --clean[-all] */
     if (clean || clean_all)
       print_clean (line);
-    else
+    /* skip for --omit-color-empty? */
+    else if (emit_colors)
       {
         /* Foreground color code is guaranteed to be set when background color code is present.  */
         if (colors[BACKGROUND] && colors[BACKGROUND]->code)
@@ -1399,16 +1463,16 @@ has_color_name (const char *str, const char *name)
 {
     char *p;
 
-    assert (strlen (str));
-    assert (strlen (name));
+    assert (strlen (str) > 0);
+    assert (strlen (name) > 0);
 
     if (!(*str == *name || *str == toupper (*name)))
       return false;
     else if (*(name + 1) != '\0'
      && !((p = strstr (str + 1, name + 1)) && p == str + 1))
       return false;
-
-    return true;
+    else
+      return true;
 }
 
 static FILE *
@@ -1430,7 +1494,7 @@ open_file (const char *file, const char *mode)
     va_start (ap, fmt);                     \
     vfprintf (stderr, fmt, ap);             \
     va_end (ap);                            \
-    fprintf (stderr, "\n");                 \
+    fprintf (stderr, "\n");
 
 static void
 vfprintf_diag (const char *fmt, ...)
